# Copyright 2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import sys

from libertine import utils


class OutputRedirector(object):
    def __init__(self, ident, cleanup=True):
        self.cleanup = cleanup

        if utils.is_snap_environment():
            self.cache_path = '%s/.cache/libertined' % os.environ['SNAP_USER_COMMON']
        else:
            self.cache_path = '%s/.cache/libertined' % os.environ['HOME']

        self.path = '{}/{}.log'.format(self.cache_path, ident)
        self._output_file = None
        self._copied_stdout = None
        self._copied_stderr = None

    def _fileno(self, file_or_fd):
        return getattr(file_or_fd, 'fileno', lambda: file_or_fd)()

    def _do_redirect(self, stream):
        fd = self._fileno(stream)
        copied_stream = os.fdopen(os.dup(fd), 'wb')
        stream.flush()
        os.dup2(self._fileno(self._output_file), fd)
        return copied_stream, fd

    def _undo_redirect(self, stream, stream_fd, copied_stream):
        stream.flush()
        os.dup2(copied_stream.fileno(), stream_fd)
        copied_stream.close()

    def __enter__(self):
        self._output_file = open(self.path, 'w')
        self._copied_stdout, self.stdout_fd = self._do_redirect(sys.stdout)
        self._copied_stderr, self.stderr_fd = self._do_redirect(sys.stderr)

        return self

    def __exit__(self, type, value, tb):
        if self._copied_stdout and self.stdout_fd:
            self._undo_redirect(sys.stdout, self.stdout_fd, self._copied_stdout)
        if self._copied_stderr and self.stderr_fd:
            self._undo_redirect(sys.stderr, self.stderr_fd, self._copied_stderr)
        if self._output_file:
            self._output_file.close()
        if self.cleanup and os.path.exists(self.path):
            os.remove(self.path)
