# Copyright 2016-2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import threading
import multiprocessing
import time
import os

from abc import ABCMeta, abstractmethod
from libertine import utils
from libertine.service import output_redirector


class BaseTask(metaclass=ABCMeta):
    """
    Abstract class for performing long-running, synchronous operations on a
    Libertine container. Child classes must implement _run, which will execute
    in a separate thread. Override _before to implement pre-execution actions
    without locking; if _before returns False, _run will not be executed.
    """
    def __init__(self, lock, container_id, config, monitor):
        self._lock = lock
        self._container = container_id
        self._config = config
        self._should_delete = multiprocessing.Event()
        self._monitor = monitor
        self._operation_id = None

    def matches(self, container, klass):
        return self._container == container and self.__class__ == klass

    @property
    def id(self):
        return self._operation_id

    @property
    def container(self):
        return self._container or ''

    @property
    def package(self):
        return ''

    @property
    def running(self):
        return not self._monitor.done(self._operation_id)

    @property
    def should_delete(self):
        return self._should_delete.is_set()

    def _finish(self):
        self._monitor.remove_from_connection(self._operation_id)
        self._should_delete.set()

    def start(self):
        self._operation_id = self._monitor.new_operation()
        self.thread = multiprocessing.Process(target=self.run)
        self.thread.start()
        return self.thread

    def watch_file(self, finished, path):
        with open(path, 'rb') as thefile:
            thefile.seek(0, 2)
            while not finished.is_set():
                finished.wait(0.1)
                line = thefile.readline().decode('utf-8')
                while line:
                    self._output(line.rstrip())
                    line = thefile.readline().decode('utf-8')

    def run(self):
        if not self._before():
            self._monitor.finished(self._operation_id)
            self._finish()
            return

        with output_redirector.OutputRedirector(self._operation_id.split("/")[-1]) as redirector:
            end_watch = multiprocessing.Event()
            multiprocessing.Process(target=self.watch_file, args=(end_watch, redirector.path)).start()
            if self._lock is not None:
                with self._lock:
                    self._run()
            else:
                self._run()

            end_watch.set()

        if self.running:
            self._monitor.finished(self._operation_id)

        self._finish()

    @abstractmethod
    def _run(self):
        pass

    def _before(self):
        return True

    def _data(self, message):
        self._monitor.data(self._operation_id, message)

    def _error(self, message):
        self._monitor.error(self._operation_id, message)

    def _output(self, message):
        self._monitor.output(self._operation_id, message)


class ContainerBaseTask(BaseTask):
    def __init__(self, lock, container_id, config, monitor, client):
        super().__init__(lock=lock, container_id=container_id, config=config, monitor=monitor)
        self._client = client
